#############################################################################
#                                                                           #
# ser2ftp.pl: Serial to FTP uploader (ver 0.1 alpha)                        #
#                                                                           #
# Copyright (C) 2007  Bueno Systems, Inc.                                   #
#                                                                           #
# Contact: Phil Pilgrim: phil@buenosystems.com                              #
#                                                                           #
# This program is free software; you can redistribute it and/or modify      #
# it under the terms of the GNU General Public License as published by      #
# the Free Software Foundation; either version 2 of the License.            #
#                                                                           #
# This program is distributed in the hope that it will be useful,           #
# but WITHOUT ANY WARRANTY; without even the implied warranty of            #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             #
# GNU General Public License for more details.                              #
#                                                                           #
# You should have received a copy of the GNU General Public License         #
# along with this program; if not, write to the Free Software               #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA #
#                                                                           #
# The GPL requires disclosure of the source code of all works derived from  #
# this one. If you wish to incorporate all or part of this program into a   #
# commercial product without disclosing the resulting source code, you may  #
# obtain a Commerical Licence (for a fee) from Bueno Systems, Inc.          #
#                                                                           #
#############################################################################

#perl2exe_include constant

use strict;
use Net::FTP;
use Win32::SerialPort qw/:STAT/;
use File::Spec;
use Cwd;
use GD;
use Time::HiRes qw/sleep/;
use Time::Local;
use Schedule::Cron::Events;
require Win32::API;

my $Port;
my $Stream = '';
my %Eols = (LF => "\n", CR => "\r", CRLF => "\r\n", LFCR => "\n\r");
my %Init = (
	maxlevel => 999,
	eol => 'LF',
	serialport => 'COM1',
	baudrate => 9600,
	ftpserver => '127.0.0.1',
	ftpuserid => 'guest',
	ftppasswd => 'guest',
	ftpdir => './',
	filedir => cwd(),
	watchdog => 120,
	resetdelay => 500
);
my ($FileName, $End, $Ext, $FileData, $SerDev, $LastReq, $XferType, $Binary, $Image, $Gd, $ColorMode, $Reset, @Palette);
my ($NPens, $NBits, $Pen, $CurX, $CurY, $PrvX, $PrvY, $JusX, $JusY, $LineSpacing, $TickType, $TickSize) = (0) x 12;
my ($Scale, $Font, $TextAngle, $Time, $TimeDate, @Time, $EventObj, $EventNext, $EventNum, $Timeout, %Time);
my $self = $0;
$self =~ s/\\/\//g;
$self =~ s/(.*)\/[^\/]*/$1/;
my $INIT_DIR = File::Spec->canonpath($self);
my $TEMP_DIR = File::Spec->tmpdir;

$| = 1;

GetInit();
$Init{eol} = $Eols{uc($Init{eol})} || "\n";
$Init{filedir} =~ s/(.*)(\/|\\)$/$1/;
$Timeout = $Init{watchdog}; 

Log(5, "Starting...");
GetTime(0);
$LastReq = 0;
while (1) {
	my $line = GetLine(0);
	Log(10, $line);
	if ($line =~ m/^\!(.*)/) {
		my $xfer;
		my @args = split(/:/, $1);
		s/\c/:/g foreach @args;
		s/\\\\/\\/g foreach @args;
		my $cmd = uc(shift @args);
		if ($cmd =~ m/^EV([1..9])$/) {
			SetEvents($args[0], $1)			
		} elsif ($cmd eq 'WDT') {
			my $t = Num($args[0]);
			$Timeout = $t;
			if ($t) {
				Log(5, "Setting watchdog timeout to $t seconds.")
			} else {
				Log(5, "Disabling watchdog timer.")
			}
		} elsif ($cmd eq 'UTC') {
			GetTime(1)
		} elsif ($cmd eq 'LTC') {
			GetTime(0)
		} elsif ($cmd eq 'ECH') {
			sleep 0.05;
			$Port->write(join(':', @args))
		} elsif ($cmd eq 'EOL') {
			$Init{eol} = $args[0] || ''			
		} elsif ($cmd eq 'FTP' or $cmd eq 'FIL') {
			$XferType = $cmd;
			$FileName = $args[0];
			$FileName =~ s/^\s*[\/\\\.]*//;
			$End = $args[-1];
			$Ext = lc((split(/\./, $FileName))[-1]) || 'txt';
			$Image = $Ext =~ m/^(jpg|gif|png)$/i;
			$Binary = $Ext =~ m/^(bin|dat)$/i;
			my $data = '';
			my $img = '';
			if ($Image) {
				($PrvX, $PrvY, $Scale) = (0, 0, 1);
				$Font = gdMediumBoldFont;
				if (uc($line) =~ m/:(RGB|G16|G256|C16|BW):/) {
					$ColorMode = $1
				} else {
					$ColorMode = 'RGB'
				}
				my ($xdim, $ydim) = (100, 100);
				if (uc($line) =~ m/:X(\d+):/) {
					$xdim = $1
				}
				if (uc($line) =~ m/:Y(\d+):/) {
					$ydim = $1
				}
				$img = GD::Image->new($xdim, $ydim, $ColorMode eq 'RGB');
				if ($ColorMode eq 'G256') {
					$Palette[$_] = $img->colorResolve($_, $_, $_) foreach 0 .. 255;
					$NPens = 256;
					$NBits = 8
				} elsif ($ColorMode eq 'G16') {
					$Palette[$_] = $img->colorResolve(($_ * 17) x 3) foreach 0 .. 15;
					$NPens = 16;
					$NBits = 4
				} elsif ($ColorMode eq 'C16') {
					$Palette[$_] = $img->colorResolve(
						(($_ & 0x04) << 5) + (($_ >> 3) & ($_ >> 2) & 1) * 127, 
						(($_ & 0x02) << 6) + (($_ >> 3) & ($_ >> 1) & 1) * 127, 
						(($_ & 0x01) << 7) + (($_ >> 3) & $_ & 1) * 127
					) foreach 0 ..6, 9.. 15;
					$Palette[7] = $img->colorResolve((0xc0) x 3);
					$Palette[8] = $img->colorResolve((0x80) x 3);
					$NPens = 16;
					$NBits = 4
				} elsif ($ColorMode eq 'BW') {
					$Palette[$_] = $img->colorResolve($_ * 255, $_ * 255, $_ * 255) foreach 0 .. 1;
					$NPens = 2;
					$NBits = 1
				} elsif ($ColorMode eq 'RGB') {
					$NPens = 0x8000000;
					$NBits = 12
				}
			}
			Log(5, "Capturing $FileName");
			while (1) {
				$line = GetLine(1);
				if ($Reset) {
					Log(1, "Timed out; aborting capture");
					last
				}
				Log(10, $line);
				if ($line eq $End) {
					if ($Image) {
						if ($Scale > 1) {
							my ($xmax, $ymax) = $img->getBounds;
							my $simg = GD::Image->new($xmax, $ymax, $ColorMode eq 'RGB');
							$simg->copyResized($img, 0, 0, 0, 0, $xmax, $ymax, $xmax / $Scale, $ymax / $Scale);
							$img = $simg;
							undef $simg
						}
						if ($Ext eq 'gif') {
							$data = $img->gif
						} elsif ($Ext eq 'jpg') {
							$data = $img->jpeg
						} elsif ($Ext eq 'png') {
							$data = $img->png
						}
						undef $img
					}
					my $file;
					if ($XferType eq 'FIL') {
						$file = File::Spec->canonpath($Init{filedir} . '/' . $FileName)
					} elsif (exists $Init{ftpserver} && exists $Init{ftpuserid} && exists $Init{ftppasswd}) {
						$file = File::Spec->canonpath($TEMP_DIR . '/ser2ftp.tmp')
					} else {
						Log(1, "Missing one or more FTP destination or login specs.");
						last
					}
					if (open(FILE, ">$file")) {
						binmode FILE;
						print FILE $data;
						close FILE;
						Log(5, "$file written.");
						if ($XferType eq 'FTP') {
							foreach my $try (0..4) {
								my $ftp;
								my $error;
								Log(5, "Logging in to FTP site $Init{ftpserver}");
								if (
									$ftp = Net::FTP->new($Init{ftpserver}, Passive => 1) and
									$ftp->login($Init{ftpuserid}, $Init{ftppasswd}) and
									$ftp->binary and
									$ftp->cwd($Init{ftpdir})
								) {
									if ($ftp->put($file, $FileName)) {
										Log(5, "File $FileName transferred");
										$ftp->quit;
										last
									} else {
										$error = 'file transfer error.'
									}
								} else {
									$error = 'login error.'
								}
								$ftp->quit if defined $ftp;
								Log(1, "FTP $error. " . ($try < 4 ? 'Retrying...' : 'Giving up.'));
								sleep 5 if $try < 4;
							}
							unlink $file
						}
					} else {
						Log(1, "Error writing to $file.")
					}
					undef $data;
					last
				} elsif ($Image) {
					my @cmdargs = $line =~ m/[ ,;]*(LB.*|[A-Z][A-Z]|[+-]?\d+|\$[\dA-F]+|%[01]+)/sgi;
					while (@cmdargs) {
						while (@cmdargs) {
							$cmd = uc shift @cmdargs;
							last if $cmd =~ m/^[A-Z][A-Z]$/
						}
						my @args = ();
						while (@cmdargs) {
							last if $cmdargs[0] =~ m/^[A-Z][A-Z]$|^LB/i;
							push @args, shift @cmdargs
						}
						Log(8, "$cmd: " . join(',', @args));
						
						if ($cmd eq 'SP') {
							
							SelectPen($img, Num($args[0]))
							
						} elsif ($cmd eq 'PC') {
							
							if ($ColorMode ne 'RGB') {
								my $pen = Num($args[0]) % $NPens;
								$img->colorDeallocate($Palette[$pen]);
								$Palette[$pen] = $img->colorResolve(Num($args[1]))
							} else {
								Log(1, "Attempt to set a palette color in RGB mode.")
							}
							
						} elsif ($cmd eq 'LT') {
							
							$img->setThickness(Num($args[0]) || 1)
							
						} elsif ($cmd eq 'MV') {
							
							GetXY(@args[0..1])
							
						} elsif ($cmd eq 'LI') {
							
							push @args, $CurX, $CurY unless @args;
							
							while (@args) {
								GetXY(shift @args, shift @args);
								$img->line($PrvX, $PrvY, $CurX, $CurY, $Pen)
							}
							
						} elsif ($cmd eq 'TT') {
							
							$TickType = Num($args[0])

						} elsif ($cmd eq 'TS') {
							
							$TickSize = Num($args[0])
							
						} elsif ($cmd eq 'TI') {
							
							push @args, $CurX, $CurY unless @args;
							
							while (@args) {
								GetXY(shift @args, shift @args);
								
								if ($TickType == 0) {
									
									$img->line($CurX - $TickSize, $CurY, $CurX + $TickSize, $CurY, $Pen)
								
								} elsif ($TickType == 1) {
								
									$img->line($CurX, $CurY - $TickSize, $CurX, $CurY + $TickSize, $Pen)
									
								} elsif ($TickType == 2) {
									
									$img->rectangle($CurX - $TickSize, $CurY - $TickSize, $CurX + $TickSize, $CurY + $TickSize, $Pen)
									
								} elsif ($TickType == 3) {
									
									$img->filledRectangle($CurX - $TickSize, $CurY - $TickSize, $CurX + $TickSize, $CurY + $TickSize, $Pen)
									
								} elsif ($TickType == 4) {
								
									$img->ellipse($CurX, $CurY, $TickSize * 2, $TickSize * 2, $Pen)
							
								} elsif ($TickType == 5) {
								
									$img->filledEllipse($CurX, $CurY, $TickSize * 2, $TickSize * 2, $Pen)
								
								}
								
							}
							
						} elsif ($cmd eq 'ER') {
							
							GetXY(@args[0..1]);
							$img->rectangle($PrvX, $PrvY, $CurX, $CurY, $Pen);
							($CurX, $CurY) = ($PrvX, $PrvY)
							
						} elsif ($cmd eq 'FR') {
							
							GetXY(@args[0..1]);
							$img->filledRectangle($PrvX, $PrvY, $CurX, $CurY, $Pen);
							($CurX, $CurY) = ($PrvX, $PrvY)
							
						} elsif ($cmd eq 'EC') {
							
							my $radius = Num($args[0]);
							$img->arc($CurX, $CurY, $radius * 2, $radius * 2, 0, 360, $Pen)
							
						} elsif ($cmd eq 'FC') {
							
							my $radius = Num($args[0]);
							$img->filledArc($CurX, $CurY, $radius * 2, $radius * 2, 0, 360, $Pen)
							
						} elsif ($cmd eq 'EW') {
							
							my ($radius, $start, $sweep) = map {Num($_)} @args[0 .. 2];
							my $end = (720 - $start % 360 - $sweep % 360) % 360;
							$start = (720 - $start % 360) % 360;
							$img->filledArc($CurX, $CurY, $radius * 2, $radius * 2, $end, $start, $Pen, gdArc | gdPie | gdNoFill | gdEdged)
							
						} elsif ($cmd eq 'FW') {
							
							my ($radius, $start, $sweep) = map {Num($_)} @args[0 .. 2];
							my $end = (720 - $start % 360 - $sweep % 360) % 360;
							$start = (720 - $start % 360) % 360;
							$img->filledArc($CurX, $CurY, $radius * 2, $radius * 2, $end, $start, $Pen)
							
						} elsif ($cmd eq 'PX') {
							
							my $radix = substr($args[0], 0, 1);
							my @pixels = split(//, substr($args[0], 1));
							next unless my $bits = $radix eq '$' ? 4 : $radix eq '%' ? 1 : 0;
							next unless my $size = int($NBits / $bits);
							while (@pixels) {
								SelectPen($img, Num($radix . join('', splice(@pixels, 0, $size))));
								$img->setPixel($CurX++, $CurY, $Pen)
							}
							
						} elsif ($cmd eq 'LD') {
							
							$TextAngle = Num($args[0]) ? 90 : 0
							
						} elsif ($cmd eq 'LS') {
							
							my $size = Num($args[0]);
							$Font = (gdTinyFont, gdSmallFont, gdMediumBoldFont, gdLargeFont, gdGiantFont)[$size > 4 ? 4 : $size]
						
						} elsif ($cmd eq 'LJ') {
							
							($JusX, $JusY, $LineSpacing) = map {Num($_)} @args[0 .. 2];
							
						} elsif ($cmd =~ m/^LB(.*)/s) {
							
							my $string = $1;
							my @strings = split(/\n/s, $string, -1);
							my $strhgt = $Font->height;
							while (@strings) {
								$string = shift @strings;
								($PrvX, $PrvY) = ($CurX, $CurY);
								my $strlen = $Font->width * length($string);
								if ($TextAngle) {
									$img->stringUp($Font, $CurX + ($JusY - 100) * $strhgt / 100, $CurY + $JusX * $strlen / 100, $string, $Pen);
									$CurY -= $strlen * (100 - $JusX) / 100
								} else {
									$img->string($Font, $CurX - $JusX * $strlen / 100, $CurY + ($JusY - 100) * $strhgt / 100, $string, $Pen);
									$CurX += $strlen * (100 - $JusX) / 100
								}
								if (@strings) {
									if ($TextAngle) {
										$CurX = $PrvX + $strhgt * ($LineSpacing + 100) / 100;
										$CurY = $PrvY
									} else {
										$CurY = $PrvY + $strhgt * ($LineSpacing + 100) / 100;
										$CurX = $PrvX
									}
								}
							}
							
						} elsif ($cmd eq 'SC') {
							
							$Scale = Num($args[1]) / (Num($args[0]) || 1);
							$Scale = 1 if $Scale < 1;
							
						} elsif ($cmd eq 'BG') {
							
							my $pen = $Pen;
							SelectPen($img, Num($args[0]));
							$img->filledRectangle(0, 0, $img->getBounds, $Pen);
							$Pen = $pen
							
						}
						
					}
				} elsif ($Binary) {
					if ($line =~ m/^([\dA-Fa-f][\dA-Fa-f])+$/) {
						foreach my $i (0 .. length($line) / 2 - 1) {
							$data .= chr(hex(substr($line, $i * 2, 2)))
						}
					}
				} else {
					$data .= $line . $Init{eol}
				}
			}
			
		} elsif ($cmd ne 'NOP') {
			Log(1, "Unrecognized command: $cmd");
		}
	}		
}

sub SelectPen {
	my ($img, $color) = @_;
	if ($ColorMode eq 'RGB') {
		$Pen = $img->colorExact(($color >> 16) & 0xff, ($color >> 8) & 0xff, $color & 0xff)
	} else {
		$Pen = $Palette[$color % $NPens]
	}
}

sub Num {
	my $value = shift;
	if (defined $value) {
		if ($value =~ m/^\$/) {
			return hex(substr($value, 1, 8))
		} elsif ($value =~ m/^%/) {
			my $ret = 0;
			foreach (1 .. length($value) - 1) {
				$ret = $ret << 1 + substr($value, $_, 1) if $_ <= 32
			}
			return $ret
		} else {
			return $value
		}
	} else {
		return 0
	}
}

sub GetXY {
	my ($x, $y) = @_;
	$PrvX = $CurX;
	$PrvY = $CurY;
	my ($newx, $newy) = (Num($x), Num($y));
	if (defined $x && defined $y) {
		$CurX = ($x =~ m/^\s*[\+\-]/ ? $CurX + $newx : $newx);
		$CurY = ($y =~ m/^\s*[\+\-]/ ? $CurY + $newy : $newy);
	}
}
	
sub GetLine {
	my $esccolon = shift;
	$Reset = 0;
	my $ntimes = 0;
	while (1) {
		sleep 60 until $Port || OpenPort();
		if (defined $EventObj && time() >= $EventNext) {
			Log(5, "Issuing scheduled event $EventNum.");
			if ($EventNum < 5) {
				ResetStamp();
				sleep $Init{resetdelay} / 1000
			}
			$Port->write("EV$EventNum");
			$EventNext = NextEvent();
		}
			
		my $nxt;
		if ($Port) {
			my (undef, $n, undef, $err) = ($Port->status);
			if ($err) {
				$Port->reset_error;
			}
			if ($n) {
				$Stream .= $Port->input;
				$LastReq = time();
				$ntimes = 0
			} elsif ($ntimes++ > 100) {
				sleep .1
			}
			if ($Stream =~ m/([^\r]*)\r/osg) {
				$nxt = pos($Stream);
				$Stream = substr($Stream, $nxt) if defined $nxt;
				my $line = $1;
				$line =~ s/\c/:/g if $esccolon;
				$line =~ s/\\q/"/g;
				$line =~ s/\\n/\n/g;
				$line =~ s/\\r/\r/g;
				$line =~ s/\\$_/$Time{$_}/g foreach keys %Time;
				$line =~ s/\\\\/\\/g if $esccolon;
				return $line
			} elsif ($Timeout && time() - $LastReq > $Timeout) {
				sleep 60 until OpenPort()
			}			
		}
	}
}

sub GetTime {
	
	Log(5, "Reading time.");
	my $t = time;
	my $gmt = shift;
	my @time = $gmt ? gmtime($t) : localtime($t);
	%Time = (
		t => $t,
		s => sprintf('%2.2d', $time[0]),
		u => sprintf('%2.2d', $time[1]),
		h => sprintf('%2.1d', $time[2]),
		H => sprintf('%2.1d', $time[2] < 1 ? 12 : ($time[2] - 1) % 12 + 1),
		p => $time[2] < 12 ? 'a.m.' : 'p.m.',
		P => $time[2] < 12 ? 'AM' : 'PM',
		d => $time[3],
		m => $time[4] + 1,
		M => (qw/Jan Feb mar Apr May Jun Jul Aug Sep Oct Nov Dec/)[$time[4]],
		y => $time[5] % 100,
		Y => $time[5] + 1900,
		w => (qw/Sun Mon Tue Wed Thu Fri Sat/)[$time[6]],
		W => (qw/Sunday Monday Tuesday Wednesday Thursday Friday Saturday/)[$time[6]],
		D => $time[7],
		Z => $gmt ? ($time[8] ? 'GDT' : 'GMT') : ($time[8] ? 'LST' : 'LDT')
	)
}	

sub SetEvents {	
	my $cron = shift;
	if (defined $cron) {
		my $Event = eval{Schedule::Cron::Events->new($cron)};
		if (ref $Event) {
			$EventObj = $Event;
			$EventNext = NextEvent();
			$EventNum = shift;
			Log(5, "Setting event $EventNum: $cron")
		} else {
			Log(1, "Can't set invalid event $EventNum: $cron")
		}
	} else {
		undef $EventObj;
		Log(5, "Clearing the event schedule.")
	}
}

sub NextEvent {	
	return defined $EventObj ? timelocal($EventObj->nextEvent) : undef
}

sub OpenPort {
	ClosePort();
	return 0 unless uc($Init{serialport}) =~ m/(COM\d\d?)/;
	$SerDev = $1;
	if ($Port = new Win32::SerialPort($1, 1)
		and $Port->baudrate($Init{baudrate})
		and $Port->parity('none')
		and $Port->databits(8)
		and $Port->stopbits(2)
		and $Port->buffers(4096,4096)
		and $Port->binary(1)
		and $Port->write_settings)
	{
		Log(5,"Port $SerDev opened.");
		ResetStamp();
		sleep $Init{resetdelay} / 1000;
		$Port->write("EV0");
		return 1
	} else {
		Log(1, "Error opening port $SerDev.");
		undef $Port;
		ClosePort();
		return 0
	}
}

sub ClosePort {
	if ($Port) {
		$Port->close;
		undef $Port;
		Log(5, "Port $SerDev closed.")
	}
}

sub ResetStamp {
	if ($Port) {
		$Port->pulse_dtr_on(2);
		$LastReq = time();
		sleep .05;
		Log(5, "Sending reset to device on $Init{serialport}");
		$Reset = 1;
		$Port->input;
		$Stream = '';
	}
}

sub Log {
	my ($level, $msg) = @_;
	$msg = '** ERROR ** ' . $msg if $level == 1;
	print scalar localtime(), ": $msg\n" if $level <= $Init{maxlevel};
	return 1
}

sub Quit {
	exit
}	

sub GetInit {
	my $ini;
	if ($ini = Slurp("$INIT_DIR/ser2ftp.ini")) {
		foreach my $line (split (/[\r\n]+/, $ini)) {
			next if $line =~ m/^\s*#/;
			if ($line =~ m/(\w+)\s*=\s*(.*)/) {
				$Init{$1} = $2
			}
		}
	}
}

sub Slurp	{
	local $/ = '';
	foreach (@_) {
		if (-e $_ && open(FILE, "<$_")) {
			binmode FILE;
			my $contents = <FILE>;
			close FILE;
			return wantarray ? ($contents, $_) : $contents
		}
	}
	return undef
}



